BIBLIOGRAPHY_STR = """

@article{haghighat2021sciann,
  title={SciANN: A Keras/TensorFlow wrapper for scientific computations and physics-informed deep learning using artificial neural networks},
  author={Haghighat, Ehsan and Juanes, Ruben},
  journal={Computer Methods in Applied Mechanics and Engineering},
  volume={373},
  pages={113552},
  year={2021},
  publisher={Elsevier},
  url={https://www.sciencedirect.com/science/article/pii/S0045782520307374}
}

@article{raissi2019physics,
  title={Physics-informed neural networks: A deep learning framework for solving forward and inverse problems involving nonlinear partial differential equations},
  author={Raissi, Maziar and Perdikaris, Paris and Karniadakis, George E},
  journal={Journal of Computational Physics},
  volume={378},
  pages={686--707},
  year={2019},
  publisher={Elsevier},
  url={https://www.sciencedirect.com/science/article/pii/S0021999118307125}
}

@article{wang2020gp,
  title={Understanding and mitigating gradient pathologies in physics-informed neural networks},
  author={Wang, Sifan and Teng, Yujun and Perdikaris, Paris},
  journal={arXiv preprint arXiv:2001.04536},
  year={2020},
  url={https://arxiv.org/abs/2001.04536}
}

@article{wang2020ntk,
  title={When and why PINNs fail to train: A neural tangent kernel perspective},
  author={Wang, Sifan and Yu, Xinling and Perdikaris, Paris},
  journal={arXiv preprint arXiv:2007.14527},
  year={2020},
  url={https://arxiv.org/abs/2007.14527}
}

@article{li2017visualizing,
  title={Visualizing the loss landscape of neural nets},
  author={Li, Hao and Xu, Zheng and Taylor, Gavin and Studer, Christoph and Goldstein, Tom},
  journal={arXiv preprint arXiv:1712.09913},
  year={2017},
  url={https://arxiv.org/abs/1712.09913}
}

@article{wang2020eigenvector,
  title={On the eigenvector bias of Fourier feature networks: From regression to solving multi-scale PDEs with physics-informed neural networks},
  author={Wang, Sifan and Wang, Hanwen and Perdikaris, Paris},
  journal={arXiv preprint arXiv:2012.10047},
  year={2020},
  url={https://arxiv.org/abs/2012.10047}
}

@article{jagtap2020locally,
  title={Locally adaptive activation functions with slope recovery for deep and physics-informed neural networks},
  author={Jagtap, Ameya D and Kawaguchi, Kenji and Em Karniadakis, George},
  journal={Proceedings of the Royal Society A},
  volume={476},
  number={2239},
  pages={20200334},
  year={2020},
  publisher={The Royal Society},
  url={https://royalsocietypublishing.org/doi/abs/10.1098/rspa.2020.0334}
}

@article{sener2018multi,
  title={Multi-task learning as multi-objective optimization},
  author={Sener, Ozan and Koltun, Vladlen},
  journal={arXiv preprint arXiv:1810.04650},
  year={2018}
}

@inproceedings{chen2018gradnorm,
  title={Gradnorm: Gradient normalization for adaptive loss balancing in deep multitask networks},
  author={Chen, Zhao and Badrinarayanan, Vijay and Lee, Chen-Yu and Rabinovich, Andrew},
  booktitle={International Conference on Machine Learning},
  pages={794--803},
  year={2018},
  organization={PMLR}
}

@article{maddu2021inversedirichlet,
      title={Inverse-Dirichlet Weighting Enables Reliable Training of Physics Informed Neural Networks},
      author={Suryanarayana Maddu and Dominik Sturm and Christian L. Müller and Ivo F. Sbalzarini},
      year={2021},
      eprint={2107.00940},
      archivePrefix={arXiv},
      primaryClass={cs.LG}
}

@article{jagtap2021deep,
  title={Deep Kronecker neural networks: A general framework for neural networks with adaptive activation functions},
  author={Jagtap, Ameya D and Shin, Yeonjong and Kawaguchi, Kenji and Karniadakis, George Em},
  journal={arXiv preprint arXiv:2105.09513},
  year={2021}
}

@article{mcclenny2020self,
  title={Self-adaptive physics-informed neural networks using a soft attention mechanism},
  author={McClenny, Levi and Braga-Neto, Ulisses},
  journal={arXiv preprint arXiv:2009.04544},
  year={2020}
}

"""